# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from requests import Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from bs4 import BeautifulSoup as bs

from functools import wraps
import errno
import os
import signal
import json 
from os.path import dirname, join
from random_proxies.proxies.log import logger
from random_proxies.proxies.exception import TimeoutError, CountryCodeError
from random_proxies.proxies import settings

def country_to_code(country, code):
    mapper = {}
    path = join(dirname(__file__), 'c2c.json')
    with open(path, 'r') as f:
        mapper = json.load(f)
    if mapper[country] == code:
        return True
    raise CountryCodeError('Country code does not match with the added country.')

# https://stackoverflow.com/questions/2281850/timeout-function-if-it-takes-too-long-to-finish
def timeout(seconds=10, error_message=os.strerror(errno.ETIME)):
    def decorator(func):
        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)
        def wrapper(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handle_timeout)
            signal.alarm(seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result
        return wraps(func)(wrapper)
    return decorator

# https://www.peterbe.com/plog/best-practice-with-retries-with-requests
def fetch(url=settings.BASE_URL,
    suffix='',
    backoff_factor=settings.HTTP_DELAY,
    status_forcelist=(500, 502, 504)):
    try:
        res = None
        with Session() as sess:
            retry = Retry(
                total=settings.HTTP_RETRIES,
                read=settings.HTTP_RETRIES,
                connect=settings.HTTP_RETRIES,
                backoff_factor=backoff_factor,
                status_forcelist=status_forcelist,
            )
            adapter = HTTPAdapter(max_retries=settings.HTTP_RETRIES)
            sess.mount('http://', adapter)
            sess.mount('https://', adapter)

            res = sess.get(url, timeout=settings.HTTP_TIMEOUT)
        return res.text
    except Exception as e:
        template = 'An exception of type {0} occurred.\nArguments: {1!r}'
        message = template.format(type(e).__name__, e.args)
        logger.error(message)
        return None

def parse_header(header):
    # Fetching field names
    fields = []
    for field in header.find_all('th'):
        name = field.get_text().strip()
        fields.append(name)
    return fields

# Do some condition checking here
def parse_values(body, fields, conditions):
    rows = body.find_all('tr')
    proxies = []
    for row in rows:
        proxy = {}
        values = row.find_all('td')
        flag = False
        for field, value in zip(fields, values):
            field = field.lower()
            value = value.get_text().strip().lower()

            # When using this routine for DB dump update
            if conditions == {}:
                proxy[field] = value
                continue

            if field == 'last checked':
                temp = field.split()
                if temp[1].startswith('second'):
                    proxy[field] = value
                    
                # Taking only proxies scanned before less than 20 minutes
                elif temp[1].startswith('minute') and int(temp[0]) < settings.LAST_CHECKED_THRESHOLD:
                    proxy[field] = value

            elif field == 'country':
                # Check if code is added or not
                if conditions.get('country')== None:
                    proxy[field] = value
                else:
                    if value.startswith(conditions.get('country')):
                        proxy[field] = value

            elif field == 'https' and conditions.get('https') == value:
                proxy[field] = value

            elif field == 'code':
                # First code must match country if both are not none
                if conditions.get('code') == None:
                    proxy[field] = value
                else:
                    if conditions.get('country') and country_to_code(conditions.get('country'), conditions.get('code')):
                        proxy[field] = value
                    elif conditions.get('code'):
                        proxy[field] = value

            elif field == 'anonymity':
                if conditions.get('anonymity') == None:
                    proxy[field] = value
                elif conditions.get('anonymity') == value:
                    proxy[field] = value

            elif field in ['ip address', 'port', 'google', 'version']:
                proxy[field] = value

            else:
                flag = True
                break
            
        if not flag:
            proxies.append(proxy)
    return proxies

# TODO
# Add to db after giving the response
# Use multiprocessing to do so
def parse_response(res, conditions):
    soup = bs(res, 'lxml')
    table = soup.find(id='proxylisttable')
    header = table.find('thead')

    fields = parse_header(header)
    body = table.find('tbody')

    proxies = parse_values(body, fields, conditions)

    return proxies



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

class TimeoutError(Exception):
    pass

class NoSuchProxyError(Exception):
    pass

class CountryCodeError(Exception):
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import requests

import random_proxies.proxies.settings as settings
from random_proxies.proxies.utils import timeout
from random_proxies.proxies.log import logger

@timeout(seconds=settings.HTTP_TIMEOUT)
def is_good_proxy(ip, protocol='http'):   
    proxies = {
        'http': 'http://' + ip,
    }
    
    if protocol == 'https':
        proxies['https'] = 'https://' + ip

    url = settings.TEST_URL
    try:
        res = requests.get(url, proxies=proxies)
        if res.status_code == 200:
            return True
        return False
    except Exception as e:
        template = 'An exception of type {0} occurred.\nArguments: {1!r}'
        message = template.format(type(e).__name__, e.args)
        logger.error(message)
        return False




# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import json
import requests

from random_proxies.proxies.settings import CACHE_SERVER_URL
from random_proxies.proxies.exception import NoSuchProxyError

def pop(conditions):
    query_string = '?'

    for k, v in conditions.items():
        query_string += f'{k}={v}&'
    url = CACHE_SERVER_URL + query_string[:-1]
    data = requests.get(url).text
    data = json.loads(data)

    if data['success'] == 'yes':
        return data['ip']
    else:
        raise NoSuchProxyError('No proxy satisfying given conditions.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import os
from os.path import join, dirname
from dotenv import load_dotenv

from pymongo import MongoClient

env_path = join(dirname(__file__), '.env')
load_dotenv(env_path)

uri = os.environ.get('MONGO_URI')

conn = MongoClient(uri)
db = conn['random_proxies']

# Remove circular import
from random_proxies.proxies.log import logger
from random_proxies.proxies.proxy_health import is_good_proxy
from random_proxies.proxies.settings import BASE_URL, SOCKS_URL, SSL_URL
from random_proxies.proxies.utils import fetch, parse_response



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s', level=logging.ERROR)

logger = logging.getLogger(__package__)


# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# Will run every day at 12 am to check for working proxies

from time import time

from random_proxies.cache import db
from random_proxies.cache import is_good_proxy
from random_proxies.cache import logger

def _clean():
    # Get all the proxies from proxies collection
    collection = db['proxies']
    proxies = collection.find({})
    
    # Delete those which arent good
    for proxy in proxies:
        ip = proxy['_id']
        protocol = ('http', 'https')[proxy['https'] == 'yes']
        
        # Implies SOCKS proxy
        if 'version' in proxy:
            ip = proxy['version'] + '://' + ip
            protocol = 'http'
        try:
            # If it doesn't work
            if not is_good_proxy(ip, protocol=protocol):
                # Delete from proxies collection
                collection.delete_one({'_id': ip})
            
        except Exception as e:
            # Delete from proxies collection
            collection.delete_one({'_id': ip})
            template = 'An exception of type {0} occurred.\nArguments: {1!r}'
            message = template.format(type(e).__name__, e.args)
            logger.error(message)

if __name__ == '__main__':
    tic = time()
    _clean()
    tac = time()
    print('Total time: [clean]', tac - tic)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

__version__ = '0.0.2'

BASE_URL = 'https://free-proxy-list.net/'
SSL_URL = 'http://sslproxies.org/'
SOCKS_URL = 'https://www.socks-proxy.net/'
TEST_URL = 'https://httpbin.org'

HTTP_DELAY = 0.2
HTTP_TIMEOUT = 6
HTTP_RETRIES = 3

LAST_CHECKED_THRESHOLD = 20
CACHE_SERVER_URL = 'http://54.162.135.16:5000/fetch'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# Will run after every 4 hours to update 'recents' collection

from time import time

from random_proxies.cache import db
from random_proxies.cache import is_good_proxy
from random_proxies.cache import logger

def _check():
    recents_collection = db['recents']
    proxies_collection = db['proxies']

    # Check if proxies are working in recents index
    recents = collection.find({})
    for proxy in recents:

        ip = proxy['_id']
        protocol = ('http', 'https')[proxy['https'] == 'yes']
        
        # Implies SOCKS proxy
        if 'version' in proxy:
            ip = proxy['version'] + '://' + ip
            protocol = 'http'

        try:
            # Only if it works
            if is_good_proxy(ip, protocol=protocol):
                # Delete from recents
                recents_collection.delete_one({'_id': ip})

                # Add them to proxies
                proxies_collection.insert_one(proxy)

        except Exception as e:
            # Delete from recents
            recents_collection.delete_one({'_id': ip})
            template = 'An exception of type {0} occurred.\nArguments: {1!r}'
            message = template.format(type(e).__name__, e.args)
            logger.error(message)
            

if __name__ == '__main__':
    tic = time()
    _check()
    tac = time()
    print('Total time: [update]', tac - tic)


# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# Will run after every 2 hours to add new proxies to 'proxies' collection

from time import time

from random_proxies.cache import db
from random_proxies.cache import fetch, parse_response
from random_proxies.cache import is_good_proxy
from random_proxies.cache import logger
from random_proxies.cache import BASE_URL, SSL_URL, SOCKS_URL    

def _check():
    urls = [BASE_URL, SSL_URL, SOCKS_URL]
    proxies = []

    # Fetch all the proxies from these urls
    for url in urls:
        res = fetch(url)
        # Passing empty conditions so that all proxies will be fetched
        proxies.extend(parse_response(res, {}))

    count = 0

    # proxies collection
    proxies_collection = db['proxies']

    # Check if they work
    for proxy in proxies:
        ip = proxy['ip address'] + ':' + proxy['port']

        # Adding _id to proxy document
        proxy['_id'] = ip
        protocol = ('http', 'https')[proxy['https'] == 'yes']

        # Implies SOCKS proxy
        if 'version' in proxy:
            ip = proxy['version'] + '://' + ip
            protocol = 'http'

        try:
            # Only if it works
            if is_good_proxy(ip, protocol=protocol):

                # Add it to proxies collection
                proxies_collection.insert_one(proxy)
                
        except Exception as e:
            template = 'An exception of type {0} occurred.\nArguments: {1!r}'
            message = template.format(type(e).__name__, e.args)
            logger.error(message)

if __name__ == '__main__':
    tic = time()
    _check()
    tac = time()
    print('Total time: [routine]', tac - tic)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from random import choice

from random_proxies.proxies import settings
from random_proxies.proxies.utils import fetch, parse_response, country_to_code
from random_proxies.proxies.exception import NoSuchProxyError
from random_proxies.proxies.db import pop

def _select(proxies):
    if len(proxies) == 0:
        raise NoSuchProxyError('No proxy satisfying given conditions.')
    proxy = choice(proxies)
    
    return proxy['ip address'] + ':' + proxy['port']

def random_proxy(
    use_cache=True,
    protocol='http',
    standard=None,
    country=None,
    code=None
):
    conditions = {
        'country': country,
        'https': ('yes', 'no')[protocol == 'http'],
        'code': code,
        'anonymity': standard
    }

    if protocol == 'http':
        url = settings.BASE_URL
    elif protocol == 'https':
        url = settings.SSL_URL
    elif protocol == 'socks':
        url = settings.SOCKS_URL
        
    if not use_cache:
        res = fetch(url)
        proxies = parse_response(res, conditions)
        return _select(proxies)
    else:
        if protocol == 'socks':
            conditions['version'] = 'socks4'

        # Removing None conditions
        new_conditions = { k:v for k, v in conditions.items() if v != None}
            
        # County-code matching
        if code != None and country != None:
            if country_to_code(country, code):
                return pop(new_conditions)

        # Fetch from db
        return pop(new_conditions)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from random_proxies.cache import db
from random_proxies.cache import logger

from random import choice
from flask import Flask, request, jsonify

app = Flask(__name__)

@app.route('/')
def index():
    return "It works!"

@app.route('/fetch', methods=['GET'])
def fetch():
    conditions = request.args
    proxies_collection = db['proxies']
    recents_collection = db['recents']

    # Fetch from proxies
    proxies = proxies_collection.find(conditions)
    if proxies.count() == 0:
        return jsonify({ 'success': 'no' })
    
    # Randomly select it
    proxies = list(proxies)
    proxy = choice(proxies)
    ip = proxy['_id']
    
    try:
        # Remove it from proxies index
        proxies_collection.delete_one({ '_id': ip })

        # Add it to recents index
        recents_collection.insert_one(proxy)

        return jsonify({ 'ip': ip, 'success': 'yes' })

    except Exception as e:
        template = 'An exception of type {0} occurred.\nArguments: {1!r}'
        message = template.format(type(e).__name__, e.args)
        logger.error(message)
        return None

if __name__ == '__main__':
    app.run(debug=True, host='0.0.0.0', port=5000)


import pathlib
from setuptools import setup, find_packages
from random_proxies.proxies.settings import __version__

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name = "random_proxies",
    version = __version__,
    description = "Get a proxy server IP on the fly!",
    long_description = README,
    long_description_content_type = "text/markdown",
    url = "https://github.com/2knal/random_proxies",
    author = "Kunal Sonawane",
    author_email = "kunal.sonawane@somaiya.edu",
    license = "MIT",
    classifiers = [
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages = find_packages(),
    include_package_data = True,
    install_requires = [
        "beautifulsoup4==4.9.0",
        "lxml==4.5.0",
        "PySocks==1.7.1",
        "python-dotenv==0.13.0",
        "requests==2.23.0"
    ]
)